/////////////////////////////////////////////
/*
*	KeyGame.java - Sample adventure game
*
*	* Author			Elpida Tzafestas
*	* Original release date		December 1997
*	* Latest release date		December 2005
*/
/////////////////////////////////////////////

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/////////////////////////////////////////////
/////////////////////////////////////////////
//
//	AdventureGame class : Subclass of Applet
//
//	This is the main abstract game class for
//	adventure games. It has default properties,
//	such as size, and defines different levels
//	of the game, canon tracking and level change
//	or restart. It uses a 2D array representing the
//	current state of the world.
//
/////////////////////////////////////////////
/////////////////////////////////////////////

abstract class AdventureGame extends JApplet implements KeyListener, MouseListener
{
static int MAX_CANONS = 3;
static int MAX_LEVEL = 3;
static int WIDTH = 30, HEIGHT = 30;

// Colors and fonts used
static Font LevelFont = new Font("TimesRoman",Font.BOLD,100);
static Color LevelColor = Color.lightGray;
static Font CanonsFont=new Font("TimesRoman",Font.ITALIC,11);
static Color CanonsColor = Color.red;
Color BackgroundColor() {return Color.yellow;}

// Offscreen image for double buffering
Image offScreenImage;

final static int OBSTACLE = 20;

int level=1, canons=MAX_CANONS;

public int world[][] = new int[WIDTH][HEIGHT];
static Point scale = new Point(12,12);
Point offset() {return new Point(0,60);}

public void initSpecial() {}
public void initWorld1() {}
public void initWorld2() {}
public void initWorld3() {}
public void doEnd() {}

public void init()
{
	initLayout();
	setSize(WIDTH * scale.x + offset().x, HEIGHT * scale.y + offset().y);
	initSpecial();
	initGame();
	initForDoubleBuffering();
	addKeyListener(this);
	addMouseListener(this);
	requestFocus();
}

public void initLayout() {}

public void initForDoubleBuffering()
{
	// Initialize offscreen image for double buffering
	offScreenImage = createImage(WIDTH * scale.x + offset().x, 
						HEIGHT * scale.y + offset().y);
}

public void paint(Graphics g) {paintDefault(g);}

public void paintDefault(Graphics g)
{
String str;
FontMetrics metrics;

Color pen = g.getColor();
Font font = g.getFont();

//	Draw borders
	g.setColor(BackgroundColor());
	g.fillRect(0,0, WIDTH * scale.x, HEIGHT * scale.y);
	g.setColor(pen);
	g.drawRect(0,0, (WIDTH * scale.x) - 1, (HEIGHT * scale.y) - 1);

//	Draw level
	g.setColor(LevelColor);
	g.setFont(LevelFont);
	metrics = g.getFontMetrics();
	str = ""+level;
	g.drawString(str,
				((WIDTH * scale.x) - metrics.stringWidth(str))/2,
				((HEIGHT * scale.y) /*- metrics.getHeight()*/)/2);

//	Draw canons
	g.setColor(CanonsColor);
	g.setFont(CanonsFont);
	metrics = g.getFontMetrics();
	str = "Canons: "+canons;
	g.drawString(str,
				(WIDTH * scale.x)-metrics.stringWidth(str)-5,
				(HEIGHT * scale.y)-metrics.getHeight()-5);

	g.setColor(pen);
	g.setFont(font);

//	The game terrain
	for (int i=0;i<WIDTH;i++)
	for (int j=0;j<HEIGHT;j++)
		drawPlace(g,i,j);
}

synchronized public void update(Graphics g)
{
	if (offScreenImage != null)
	{
		Graphics offScreenGraphics = offScreenImage.getGraphics();

		paint(offScreenGraphics);
		g.drawImage(offScreenImage,0,0,this);
	}
	else paint(g);
}

void initGame()
{
	level = 1;
	canons = MAX_CANONS;
	initWorld(level);
}

public void initWorld(int lev)
{
	switch (lev)
	{
		case 1: initWorld1(); break;
		case 2: initWorld2(); break;
		case 3: initWorld3(); break;
		default: initWorld3(); break;
	}
}

void initDefault()
{
	for (int i=0;i<WIDTH;i++)
	for (int j=0;j<HEIGHT;j++)
		world[i][j] = 0;
}

void gotoNextLevel()
{
	if (level == MAX_LEVEL) doEnd();
	else
	{
		preGotoNextLevel();
		level++;
		initWorld(level);
	}
	repaint();
}

void preGotoNextLevel() {}

void restartLevel()
{
	initWorld(level);
	repaint();
}

public void drawPlace(Graphics g, int x, int y) {}
public boolean runnable() {return (canons > 0);}

boolean freePlace(int x, int y)
{
	if ((x<0) || (x>=WIDTH) || (y<0) || (y>=HEIGHT)) return false;
	return (world[x][y] == 0);
}

boolean validPlace(int x, int y)
{
	if ((x<0) || (x>=WIDTH) || (y<0) || (y>=HEIGHT)) return false;
	return true;
}

/////////////////////////////////////////////
//		EVENT HANDLERS
/////////////////////////////////////////////

public void keyReleased(KeyEvent e) {}
public void keyTyped(KeyEvent e) {}

public void mouseEntered(MouseEvent e) {}
public void mouseExited(MouseEvent e) {}
public void mouseClicked(MouseEvent e) {}
public void mouseReleased(MouseEvent e) {}

}	// End AdventureGame

/////////////////////////////////////////////
/////////////////////////////////////////////
//
//	AdventureGameAgent class
//
//	This is the main abstract agent class for
//	adventure games. It is endowed by default
//	with the possibility to move around and to do
//	something special on encounter of non-empty
//	places.
//
/////////////////////////////////////////////
/////////////////////////////////////////////

abstract class AdventureGameAgent
{
AdventureGame game;

Point position = new Point(1,1);
int heading = KeyEvent.VK_RIGHT;

abstract Color color();
abstract public void putAt(int x, int y);
abstract public void removeFrom(int x, int y);
abstract public void doSpecialAt(int x, int y);
Color headingColor() {return null;}

AdventureGameAgent(AdventureGame g)
{
	game = g;
}

public void init(int x, int y)
{
	position.x = x;
	position.y = y;
}

public void draw(Graphics g)
{
	Color pen = g.getColor();
	Color c = headingColor();

	g.setColor(color());
	g.fillOval(position.x*game.scale.x, position.y*game.scale.y, game.scale.x, game.scale.y);
	if (c != null) g.setColor(c);
		else g.setColor(pen);
	int angle = computeHeadingStartAngle();
	g.fillArc(position.x*game.scale.x, position.y*game.scale.y, game.scale.x, game.scale.y, angle, 70);
	g.setColor(pen);
}

int computeHeadingStartAngle()
{
	switch (heading)
	{
		case KeyEvent.VK_UP:
			return 55;
		case KeyEvent.VK_DOWN:
			return -125;
		case KeyEvent.VK_LEFT:
			return 145;
		case KeyEvent.VK_RIGHT:
			return -35;
		default:
			return 60;
	}
}

public void moveBy(int direction)
{
	heading = direction;
	game.repaint();
	Point step = getStepForHeading();
	Point targetPosition = new Point(position.x + step.x, position.y + step.y);
	if (!game.validPlace(targetPosition.x,targetPosition.y)) return;
	if (game.freePlace(targetPosition.x,targetPosition.y))
	{
		gotoposition(targetPosition.x,targetPosition.y);
	}
	// Do something special when the target place is not empty
	else doSpecialAt(targetPosition.x,targetPosition.y);
}

public void setPosition(int x, int y)
{
	position.x = x;
	position.y = y;
}

public void gotoposition(int x, int y)
{
	removeFrom(position.x,position.y);
	setPosition(x,y);
	putAt(x,y);
}

Point getStepForHeading()
{
	switch (heading)
	{
		case KeyEvent.VK_UP: return new Point(0,-1);
		case KeyEvent.VK_DOWN: return new Point(0,1);
		case KeyEvent.VK_RIGHT: return new Point(1,0);
		case KeyEvent.VK_LEFT: return new Point(-1,0);
		default: return new Point(0,-1);
	}
}

Point nextPlace()
{
	switch (heading)
	{
		case KeyEvent.VK_UP: return new Point(position.x,position.y-1);
		case KeyEvent.VK_DOWN: return new Point(position.x,position.y+1);
		case KeyEvent.VK_RIGHT: return new Point(position.x+1,position.y);
		case KeyEvent.VK_LEFT: return new Point(position.x-1,position.y);
		default: return new Point(position.x,position.y-1);
	}
}

int findDirectionTo(int x, int y)
	{return findCorrectDirectionTo(x,y);}

int findPerturbedDirectionTo(int x, int y, double probability)
{
//	Find the direction to a given target point (x,y).
//	With a given probability, return a random direction

	if (Math.random() < probability)
	{
		switch ((int)(Math.round((Math.random() * 4) - 0.5)))
		{
			case 0: return KeyEvent.VK_DOWN;
			case 1: return KeyEvent.VK_UP;
			case 2: return KeyEvent.VK_RIGHT;
			case 3: return KeyEvent.VK_LEFT;
			default: return KeyEvent.VK_DOWN;
		}
	}
	else return findCorrectDirectionTo(x,y);
}

int findCorrectDirectionTo(int x, int y)
{
//	Find the direction to a given target point (x,y).

	int dx = x - position.x;
	int dy = y - position.y;

	if (Math.abs(dx) < Math.abs(dy))
		return ( (dy>0) ? KeyEvent.VK_DOWN : KeyEvent.VK_UP );
	else
		return ( (dx>0) ? KeyEvent.VK_RIGHT : KeyEvent.VK_LEFT );
}

}	// End AdventureGameAgent

/////////////////////////////////////////////
/////////////////////////////////////////////
//
//	Enemy class : Subclass of AdventureGameAgent
//
//	This is the main abstract enemy class for
//	adventure games.
//
/////////////////////////////////////////////
/////////////////////////////////////////////

abstract class Enemy extends AdventureGameAgent implements Runnable
{
static Color enemyColor = Color.red;
static Color deadColor = Color.white;
Thread engine;
int sleepTime = 1000;
boolean threadControlVariable = false;

Enemy(AdventureGame g)
{
	super(g);
}

public void init(int x, int y, int s)
{
	super.init(x,y);
	sleepTime = s;
}

public Color color()
{
	if (isAlive()) return enemyColor;
	else return deadColor;
}

public void run()
{
	while (game.runnable())
	{
		if (threadControlVariable)
		{
			doRun();
			game.repaint();
		}
		try {engine.sleep(sleepTime);}
		catch (InterruptedException e) {}
	}
}

abstract public void doRun();
public void start()
{
	if (engine == null)
	{
		engine = new Thread(this);
		engine.start();
		threadControlVariable = true;
	}
	else if (!engine.isAlive())
	{
		engine = new Thread(this);
		engine.start();
		threadControlVariable = true;
	}
}

public void kill()
{
	if (engine != null)
	{
		if (engine.isAlive()) threadControlVariable = false;
		engine = null;
	}
}

public boolean isAlive()
{
	if (engine == null) return false;
	return (engine.isAlive());
}

}	// End Enemy

/////////////////////////////////////////////
/////////////////////////////////////////////
//
//	KeyGame class : Subclass of AdventureGame
//
//	This game involves a hero moving around
//	in space, whose goal is to escape its enemy
//	through a door. To open the door, it has to
//	obtain a key, that is initially located in some
//	place of the world.
//
/////////////////////////////////////////////
/////////////////////////////////////////////

public class KeyGame extends AdventureGame implements Runnable
{
//	Colors and fonts used
static Color ObstacleColor = Color.darkGray;
static Color DoorBodyColor = Color.magenta.darker();
static Color DoorHandleColor = Color.yellow;
static Color KeyColor = Color.blue;
static Font GameOverFont=new Font("TimesRoman",Font.BOLD,60);
static Color GameOverColor = Color.red;
static Font CongratulationsFont=new Font("TimesRoman",Font.BOLD,24);
static Color CongratulationsColor = Color.blue;
static Font RestartFont=new Font("TimesRoman",Font.ITALIC,12);
//Color BackgroundColor() {return Color.cyan;}
Color BackgroundColor() {return new Color(255,251,196);}

//	Hero and monster of the game
KeyHero hero;
KeyMonster monster;

//	Variables used to delay before restarting the current level
private boolean delaySameLevel = false;
static Font DelaySameLevelFont = new Font("TimesRoman",Font.BOLD,24);
static Color DelaySameLevelColor = Color.pink;
static String DelaySameLevelMessage = "Oooops!";

Thread delayEngine = null;
static long delayTime = 3000;

//	Variables used to delay before going to the next level
private boolean delayNextLevel = false;
static Font DelayNextLevelFont=new Font("TimesRoman",Font.BOLD,24);
static Color DelayNextLevelColor = Color.blue;
static String DelayNextLevelMessage = "Well done!";

//	Type constants
final static int HERO = 5;
final static int KEY = 6;
final static int DOOR = 7;
final static int MONSTER = 8;

Point offset() {return new Point(0,80);}

public void initSpecial()
{
	hero = new KeyHero(this);
	monster = new KeyMonster(this);
	delayEngine = new Thread(this);
	delayEngine.start();
}

public void initWorld1()
{
	int i,j;

	initDefault();
//	Init hero
	hero.init(1,29);
	world[1][29] = HERO;
//	Init monster
	monster.init(1,1,120);
	world[1][1] = MONSTER;
//	Init obstacles
	for (j=5;j<=25;j+=5)
	{
		for (i=3;i<12;i++)
		{
			world[i][j] = OBSTACLE;
			world[14+i][j] = OBSTACLE;
		}
	}
	for (i=1;i<5;i++) world[i][15] = OBSTACLE;
	for (j=15;j<=20;j++) world[25][j] = OBSTACLE;
	world[17][11] = OBSTACLE; world[17][12] = OBSTACLE; world[17][13] = OBSTACLE;
//	Init key
	world[14][15] = KEY;
//	Init door
	world[29][14] = DOOR;
}

public void initWorld2()
{
	int i,j;

	initDefault();
//	Init hero
	hero.init(1,29);
	world[1][29] = HERO;
//	Init monster
	monster.init(14,15,100);
	world[14][15] = MONSTER;
//	Init obstacles
	for (j=5;j<=25;j+=5)
	{
		for (i=3;i<12;i++)
		{
			world[i][j] = OBSTACLE;
			world[14+i][j] = OBSTACLE;
		}
		world[2][j] = OBSTACLE;
	}
	world[0][5] = OBSTACLE;
	world[0][25] = OBSTACLE;
	world[1][5] = OBSTACLE;
	world[1][25] = OBSTACLE;
	for (j=10;j<=20;j++) world[25][j] = OBSTACLE;
	for (i=26;i<=29;i++)
		{world[i][5] = OBSTACLE; world[i][25] = OBSTACLE;}
	for (j=5;j<9;j++) world[17][j] = OBSTACLE;
//	Init key
	world[1][1] = KEY;
//	Init door
	world[29][14] = DOOR;
}

public void initWorld3()
{
	int i,j;

	initDefault();
//	Init hero
	hero.init(1,29);
	world[1][29] = HERO;
//	Init monster
	monster.init(14,15,80);
	world[14][15] = MONSTER;
//	Init obstacles
	for (j=5;j<=25;j+=5)
	{
		for (i=3;i<12;i++)
		{
			world[i][j] = OBSTACLE;
			world[14+i][j] = OBSTACLE;
		}
		world[2][j] = OBSTACLE;
	}
	world[2][25] = OBSTACLE;
	world[0][15] = OBSTACLE; world[1][15] = OBSTACLE; world[2][15] = OBSTACLE;
	for (j=10;j<=20;j++)
	{
		world[5][j] = OBSTACLE;
		world[25][j] = OBSTACLE;
	}
	for (j=6;j<=9;j++)
		world[3][j] = OBSTACLE;
	for (j=0;j<=4;j++)
		world[25][j] = OBSTACLE;
	for (i=26;i<=29;i++)
		world[i][20] = OBSTACLE;
//	Init key
	world[24][18] = KEY;
//	Init door
	world[14][29] = DOOR;
}

void initGame()
{
	super.initGame();
	monster.start();
}

public void doEnd()
{
	hero.won = true;
}

public boolean runnable()
{
	return (hero.alive && (!hero.won));
}

public boolean paused() {return (delaySameLevel || delayNextLevel);}

/////////////////////////////////////////////
//	Event handlers
/////////////////////////////////////////////

public void keyPressed(KeyEvent evt)
{
	char key = evt.getKeyChar();
	int keycode = evt.getKeyCode();

	if (delaySameLevel || delayNextLevel) return;

	if ((!hero.alive) || (hero.won))
	{
	//	Press 's' or 'S' to restart
		if ((key == 's') || (key == 'S'))
		{
			init();
			repaint();
		}
	}

	//	Press 'g' or 'G' to grasp the key
	else if ((key == 'g') || (key == 'G'))
	{
		hero.graspKey();
		repaint();
	}

	//	Press 'd' or 'D' to drop the key
	else if ((key == 'd') || (key == 'D'))
	{
		hero.dropKey();
		repaint();
	}

	//	Press 'b' or 'B' to unlock the door
	else if ((key == 'b') || (key == 'B'))
	{
		hero.openDoor();
		repaint();
	}

	else switch (keycode)
	{
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_UP:
			hero.moveBy(keycode);
			repaint();
			break;
		default:
			break;
	}
	return;
}

public void mousePressed(MouseEvent evt)
{
	if ((!hero.alive) || (hero.won))
	{
		initGame();
		repaint();
	}
}

public void paint(Graphics g)
{
	Color pen = g.getColor();
	Font font = g.getFont();
	FontMetrics metrics;
	String str;

	paintDefault(g);
	hero.draw(g);
	monster.draw(g);

	//	Check hero dead
	if (!hero.alive)
	{
		g.setColor(GameOverColor);
		g.setFont(GameOverFont);
		metrics = g.getFontMetrics();
		str = "Game Over";
		g.drawString(str,
				((WIDTH * scale.x)-metrics.stringWidth(str))/2,
				((HEIGHT * scale.y)-metrics.getHeight())/2);
		g.setFont(CongratulationsFont);
		g.setFont(RestartFont);
		metrics = g.getFontMetrics();
		str = "Press 'S' or click to restart";
		g.drawString(str,
				((WIDTH * scale.x)-metrics.stringWidth(str))/2,
				(HEIGHT * scale.y)/2+metrics.getHeight());
		g.setColor(pen);
		g.setFont(font);
	}

	//	Check hero won
	else if (hero.won)
	{
		g.setColor(CongratulationsColor);
		g.setFont(CongratulationsFont);
		metrics = g.getFontMetrics();
		str = "Congratulations ! You won !";
		g.drawString(str,
				((WIDTH * scale.x)-metrics.stringWidth(str))/2,
				(HEIGHT * scale.y)/2);
		g.setFont(RestartFont);
		metrics = g.getFontMetrics();
		str = "Press 'S' or click to restart";
		g.drawString(str,
				((WIDTH * scale.x)-metrics.stringWidth(str))/2,
				(HEIGHT * scale.y)/2+metrics.getHeight());
		g.setColor(pen);
		g.setFont(font);
	}

	//	Check delay before restarting the current level
	else if (delaySameLevel)
	{
		g.setColor(DelaySameLevelColor);
		g.setFont(DelaySameLevelFont);
		metrics = g.getFontMetrics();
		str = DelaySameLevelMessage;
		g.drawString(str,
				((WIDTH * scale.x)-metrics.stringWidth(str))/2,
				((HEIGHT * scale.y)-metrics.getHeight())/2);
		g.setColor(pen);
		g.setFont(font);
	}

	//	Check delay before going to the next level
	else if (delayNextLevel)
	{
		g.setColor(DelayNextLevelColor);
		g.setFont(DelayNextLevelFont);
		metrics = g.getFontMetrics();
		str = DelayNextLevelMessage;
		g.drawString(str,
				((WIDTH * scale.x)-metrics.stringWidth(str))/2,
				((HEIGHT * scale.y)-metrics.getHeight())/2);
		g.setColor(pen);
		g.setFont(font);
		g.drawString("What?",100,100);
	}
}

public void drawPlace(Graphics g, int x, int y)
{
	switch (world[x][y])
	{
		case OBSTACLE: drawObstacle(g,x,y); break;
		case KEY: drawKey(g,x,y); break;
		case DOOR: drawDoor(g,x,y); break;
		default: break;
	}
}

void drawObstacle(Graphics g, int x, int y)
{
//	Dark gray block representation
	Color pen = g.getColor();

	g.setColor(ObstacleColor);
	g.fillRoundRect(x*scale.x, y*scale.y, scale.x, scale.y, 2, 2);
	g.setColor(pen);
	g.drawRoundRect(x*scale.x, y*scale.y, scale.x, scale.y, 2, 2);
}

void drawKey(Graphics g, int x, int y)
{
//	Primitive key representation
	Color pen = g.getColor();

	g.setColor(KeyColor);
	g.drawRect(x*scale.x+2,y*scale.y,5,scale.y/3);
	g.drawRect(x*scale.x+3,y*scale.y,3,scale.y/3);
	g.fillRect(x*scale.x+2,y*scale.y,2,scale.y);
	g.fillRect(x*scale.x+2,((y+1)*scale.x)-3,5,3);
	g.setColor(pen);
}

void drawDoor(Graphics g, int x, int y)
{
//	Dark magenta door with a golden handle
	Color pen = g.getColor();

	int doorWidth = (int)(scale.x*0.8);
	g.setColor(DoorBodyColor);
	g.fillRect(x*scale.x, y*scale.y, doorWidth, scale.y);
	g.setColor(DoorHandleColor);
	g.fillOval(x*scale.x+doorWidth-4,y*scale.y+(scale.y/2)-2,4,4);
	g.setColor(pen);
}

/////////////////////////////////////////////
//	"Action" methods : retrieving and setting
//	the contents of a place in the world.
/////////////////////////////////////////////

boolean keyAt(int x, int y)
{
	if (validPlace(x,y)) return (world[x][y] == KEY);
	return false;
}

boolean doorAt(int x, int y)
{
	if (validPlace(x,y)) return (world[x][y] == DOOR);
	return false;
}

boolean monsterAt(int x, int y)
{
	if (validPlace(x,y)) return (world[x][y] == MONSTER);
	return false;
}

boolean heroAt(int x, int y)
{
	if (validPlace(x,y)) return (world[x][y] == HERO);
	return false;
}

boolean validPlace(int x, int y)
{
	if (!super.validPlace(x,y)) return false;
	return (world[x][y] != OBSTACLE);
}

void putKeyAt(int x, int y)
{
	world[x][y] = KEY;
}

void putHeroAt(int x, int y)
{
	world[x][y] = HERO;
}

void putMonsterAt(int x, int y)
{
	world[x][y] = MONSTER;
}

void removeAt(int x, int y)
{
	world[x][y] = 0;
}

public void killHero()
{
	canons--;
	if (canons == 0)
	{
		hero.alive = false;
		repaint();
	}
	else
	{
		doDelaySameLevel();
		repaint();
		restartLevel();
	}
}

void doDelaySameLevel()
{
	delaySameLevel = true;
	repaint();
}

void doDelayNextLevel()
{
	delayNextLevel = true;
	repaint();
}


void preGotoNextLevel()
{
	doDelayNextLevel();
	repaint();
}

public void run()
{
	while (true)
	{
		if (delaySameLevel)
		{
			try {delayEngine.sleep(delayTime);}
			catch (InterruptedException e) {}
			delaySameLevel = false;
		}
		else if (delayNextLevel)
		{
			try {delayEngine.sleep(delayTime);}
			catch (InterruptedException e) {}
			delayNextLevel = false;
		}
		else
			try {delayEngine.sleep(10);}
			catch (InterruptedException e) {}
	}
}

}	// End KeyGame

/////////////////////////////////////////////
/////////////////////////////////////////////
//
//	KeyHero class : Subclass of AdventureGameAgent
//
//	This is the hero of the "Key Game".
//	His goal is to escape its enemy
//	through a door. To open the door, he has to
//	obtain a key, that is initially located in some
//	place of the world. The hero may move with
//	the arrows, grasp the key with the 'g' or 'G'
//	key when he is next to it, drop it to its facing
//	place with 'd' or 'D' and open the door with
//	'b' or 'B' when he has the key.
//
/////////////////////////////////////////////
/////////////////////////////////////////////

class KeyHero extends AdventureGameAgent
{
static Color heroColor = Color.green;

boolean hasKey = false;
boolean alive = true;
boolean levelDone = false;
boolean won = false;

KeyHero(KeyGame g)
{
	super((AdventureGame)g);
}

public Color color()
{
	if (hasKey) return ((KeyGame)game).KeyColor;
	if (levelDone) return Color.yellow;
	return heroColor;
}

public Color headingColor()
{
	if (hasKey) return Color.pink;
	return super.headingColor();
}

public void removeFrom(int x, int y)
	{((KeyGame)game).removeAt(x,y);}

public void putAt(int x, int y)
	{((KeyGame)game).putHeroAt(x,y);}

public void init(int x, int y)
{
	super.init(x,y);
	hasKey = false;
	alive = true;
	levelDone = false;
	won = false;
}

public void graspKey()
{
	Point target = nextPlace();
	if (((KeyGame)game).keyAt(target.x,target.y))
	{
		hasKey = true;
		((KeyGame)game).removeAt(target.x,target.y);
	}
}

public void dropKey()
{
	if (!hasKey) return;
	Point target = nextPlace();
	if (game.freePlace(target.x,target.y))
	{
		hasKey = false;
		((KeyGame)game).putKeyAt(target.x,target.y);
	}
}

public void openDoor()
{
	Point target = nextPlace();
	if (hasKey && (((KeyGame)game).doorAt(target.x,target.y)))
	{
		levelDone = true;
		game.repaint();
		game.gotoNextLevel();
	}
}

public void doSpecialAt(int x, int y)
{
	if (((KeyGame)game).monsterAt(x,y))
	{
		gotoposition(x,y);
		((KeyGame)game).killHero();
	}
}

}	// End KeyHero

/////////////////////////////////////////////
/////////////////////////////////////////////
//
//	KeyMonster class : Subclass of Enemy
//
//	This is the monster of the "Key Game".
//	His goal is to chase and kill the hero of
//	the game..
//
/////////////////////////////////////////////
/////////////////////////////////////////////

class KeyMonster extends Enemy
{
KeyMonster(KeyGame g)
{
	super((AdventureGame)g);
}

public void removeFrom(int x, int y)
	{((KeyGame)game).removeAt(x,y);}

public void putAt(int x, int y)
	{((KeyGame)game).putMonsterAt(x,y);}

public void doSpecialAt(int x, int y)
{
	if (((KeyGame)game).heroAt(x,y))
	{
		gotoposition(x,y);
		((KeyGame)game).killHero();
	}
}

public void doRun()
{
//	Chase hero.
	if (((KeyGame)game).paused()) return;
	int dir = findDirectionTo(
				((KeyGame)game).hero.position.x,
				((KeyGame)game).hero.position.y);
	moveBy(dir);
}

int findDirectionTo(int x, int y)
	{return findPerturbedDirectionTo(x,y,0.1);}

}	// End KeyMonster

